package tfgMarriageStable;


import java.io.IOException;
import java.io.Serializable;
import jade.content.lang.sl.SLCodec;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPANames;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.NotUnderstoodException;
import jade.domain.FIPAAgentManagement.Property;
import jade.domain.FIPAAgentManagement.SearchConstraints;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.lang.acl.UnreadableException;
import jade.proto.AchieveREInitiator;
import jade.proto.AchieveREResponder;
import jade.util.leap.ArrayList;


/**
 * <p>Clase que representa los diferentes elementos que se emparejarán.</p>
 * <p>Al iniciarse recibe como argumentos:</p>
 * <ul> * 
 * <li>El número de parejas esperadas</li>
 * <li>El número de características que tendrá  </li>
 * <li>La iteración en la que se está creando</li>
 * </ul> 
 * <p>Genera aleatoriamente los valores de las características y unas preferencias para cada una de ellas</p>
 * <p>Es capaz de recibir como mensaje una lista de las parejas del otro grupo.</p>
 * <p>Envía un mensaje a cada una de esas potenciales parejas preguntándole sus características
 * Cuando las recibe calcula la valoración de la pareja y reordena su lista de parejas según esas valoraciones, de mayor a menor</p>
 * <p>Una vez recibidas todas las valoraciones envía un mensaje a los emparejadores con su AID, sus características, sus preferencias y su lista de parejas ordenada
 * 
 * @author Alberto Hueso Alonso
 *@since 06-12-2017
 */
public abstract  class Elemento extends Agent {
	
	/**
	 * Constante para serializar
	 */
	private static final long serialVersionUID = -53557412590521075L;


	//Constantes
	/**Número máximo de intentos que buscará parejas, encontrando el mismo número de parejas sin llegar al esperado*/
	static final int NUMERO_INTENTOS_BUSQUEDAS=5;   
	
	
	/**Número de parejas que se esperan*/
	private int numeroParejas=0;
	/**Numero de caracteristicas que tiene el elemento*/
	protected int numeroCaracteristicas=1;
	
	/**Guarda la lista de caracteristicas del elemento*/
	private jade.util.leap.ArrayList caracteristicas;
	
	/**Guarda la lista de preferencias por cada característica del elemento*/
	 private jade.util.leap.ArrayList preferencias;
		
	/**Guarda la lista de agentes con los que se puede relacionar como pareja	*/
	private ListaParejas listaParejas=new ListaParejas();
	
	/**Guarda la lista de emparejadores encontrados*/
	private jade.util.leap.ArrayList listaEmparejadores=new ArrayList();
	
	//Variables de control	
	
	/**Guarda a cuántos agentes se ha preguntado las características*/		
	private int agentesPreguntados=0; 
	/**Guarda el número de valoraciones que ha realizado el elemento*/
	private int numeroValoraciones=0;
	/**Guarda a qué iteración de la creación de agentes pertenece al elemento*/
	private int iteracionActual=1;
	
	/**Guarda si las preferencias estarán moduladas o serán totalmente aleatorias*/	
	protected boolean moduladas=false;

	/**
	 * Guarda el AID del gestor de agentes que creó este elemento
	 */
	private AID gestorAgentes;
	
	/**
		 * <p>Inicialización del agente Elemento</p>
		 * <p>Recupera como argumento el número de parejas que tendrá</p>
		 * <p>Recupera como argumento el número de características y las iniciará</p>
		 * <p>Recupera como argumento la iteración a la que pertence el agente</p>
		 * <p>Recupera como argumento si el elemento modulará sus preferencias o no
		 * <p>Registra el agente en el DF</p>
		 * <p>Registra el lenguaje empleado,SL</p>
		 * <p>Abre escuchadores, manejadores y envía mensajes para:
		 * <ul>
		 * <li>Recibir su lista de parejas.</li>
		 * <li>Preguntar a su lista de parejas las características de cada uno.</li>
		 * <li>Responder a otros agentes que pregunten por su lista de características.</li>
		 * </ul>
		 * </p>
		 * @see #iniciarCaracteristicas()
		 * @see #iniciarPreferencias()
		 * @see #RegistrarDF()
		 * @see Funciones#registrarLenguajeSL(Agent)
		 * @see EscucharObjetoConversation
		 * @see ResponderQuery 
		 * @see EnviarMensajeAParejas
		 * @see #crearPlantillaQuery()
		 * @see #crearQuery(String)
		 * 
		 */
	    public  void setup() {
	    	
	    	//Inicializamos los atributos
	    			    
		    numeroParejas=(int) getArguments()[0];//Recuperamos el número de parejas
		    
		    numeroCaracteristicas=(int) getArguments()[1]; //Recuperamos el número de características;
		    
		    iteracionActual=(int) getArguments()[2];	//Recuperamos la iteración actual
		    
		    moduladas=(boolean) getArguments()[3]; //Recuperamos si las preferencias serán moduladas o no
		    
		    gestorAgentes=(AID) getArguments()[4]; //Recuperamos el AID del gestor de agentes
		    //Iniciamos caracteristicas y preferencias
		    this.caracteristicas=this.iniciarCaracteristicas();
		    this.preferencias=this.iniciarPreferencias();		   
		    
		    //Registramos el elemento
		    this.RegistrarDF();
		        
	   	
	   	//Registramos el lenguaje SL en el agente
	   	Funciones.registrarLenguajeSL(this);   	
	   	
	   
	    System.out.println("El agente se ha iniciado: "+getName());
	   
	    //Añadimos comportamiento para recibir nuestra lista de parejas
	    addBehaviour(new EscucharObjetoConversation("listaParejas"));
	    
	  
			
		    //Plantilla para responder queries (protocolo FIPA-QUERY)
		    MessageTemplate plantilla = crearPlantillaQuery();
	
		    //Añadimos comportamiento para responder a queries que cumplan esa plantilla  
		    this.addBehaviour(new ResponderQuery(this, plantilla));
	
		    //Creamos el mensaje de la consulta	        
		    ACLMessage mensaje = crearQuery("dimeTusCaracteristicas");	        
	
		    addBehaviour(new EnviarMensajeAParejas(mensaje));
        
	   
    }
	    /**
	     * <p>Registra el agente en el Domain Facilitator</p>
	     * <p>Asigna como nombre del servicio "conocerParejasX" o "conocerParejasY", según el agente sea Y o X respectivamente</p>
	     *<p>Asigna como tipo "parejas" seguido del número de iteración en curso</p>
	     *Asigna como propietario "TFG"
	     */
	   private void RegistrarDF(){
	    	
	    	
	    	//Petición de registro en el DF
	    	DFAgentDescription dfd=new DFAgentDescription();
	    	
	    	//Servicio que se proporciona
	    	ServiceDescription sd= new ServiceDescription();
	    	
	    	//Si es de tipo X se ofrece conocerParejasY
	    	if (this.getClass().toString().contains("X")){
	    	sd.setName("conocer_parejasY");
	    	
	    	}
	    	//Si es de tipo Y se ofrece conocerParejasX
	    	if (this.getClass().toString().contains("Y")) { 	
	    	sd.setName("conocer_parejasX");
	    	
	    	}
	    	//Se añade el tipo y el propietario
	    	String tipo="parejas"+iteracionActual;
	    	sd.setType(tipo);
	    	sd.setOwnership("TFG");
	    	
	    	//Creamos las propiedades
	    	Property propiedad= new Property();
	 
	    	
	    	//Damos valor a las propiedades
	    	propiedad.setName("propiedad");
	    	propiedad.setValue("valorPropiedad");	    	
	    	
	    
	    	//Añadimos las propiedades a los servicios	
	    	sd.addProperties(propiedad);
	    	
	    	//Añadimos los servicios al descriptor de agentes
	    	dfd.addServices(sd);
	    	dfd.setName(getAID());
	    	
	    	try {
	    		//Añadimos al registro
				DFService.register(this, dfd);
				System.out.println("Se ha registrado el agente "+getLocalName());
			} catch (FIPAException e) {				
				
					System.err.println("No se ha podido registrar el agente "+getLocalName());
					
					
							
			}
	    	
	    }
	    
	/**
	 * <p>Libera recursos y desregistra el agente</p>
	 * @see #deRegistrarDF
	 */
    protected void takeDown() { // liberación de recursos del agente
    	deRegistrarDF();	
    	
    	
    }
    
    
    
    
    /**
     * <p>Desregistra el el elemento del Domain Facilitator</p>
     */
   private void deRegistrarDF(){
    	
    try {
    	//Solicitamos el desregistro
		DFService.deregister(this);
		System.out.println("El agente "+getLocalName()+" se ha desregistrado");
	} catch (FIPAException e) {
		
		System.err.println("No se ha podido desregistrar el agente");
	}
   }
    
    /**
     * <p>Función que crea las características del elemento</p>
     * <p>Hasta igualar el número de características recibidas como argumento
     *  añade a una lista de Caracteristicas
     * el nombre de una característica ("caracteristica"+Int)
     * junto con el valor calculado como un número entre 0 y 1 </p>   
     * @return ArrayList con la lista de valores de las características que tiene el agente
     * @see Funciones#getDecimal
     * @see Caracteristica
     */
    private jade.util.leap.ArrayList iniciarCaracteristicas(){
    	
    	 jade.util.leap.ArrayList caracteristicas=new ArrayList();
       	 int i=0;
    	 while (i<this.numeroCaracteristicas){
    		 
    		 String nombre="caracteristica";
    		 nombre=nombre+i;
    		 float  valor=Funciones.getDecimal();//Le damos un valor aleatorio
    		 Caracteristica caracteristica=new Caracteristica(nombre,valor);
    		 caracteristicas.add(caracteristica);
    		 i++;
    		 
    		 //Control
    		 System.out.println(this.getLocalName()+": "+nombre+" "+valor);
    	 }
    	 
    	 
    	 
    	 		return caracteristicas;
    }
    	
    
    
    
    
    
    
    /**
     * <p>Función que crea las preferencias del elemento</p>
     * <p>Hasta igualar el número de características añade a una lista de Caracteristicas
     * el nombre de una preferencia ("caracteristica"+numero+Preferencia)
     * junto con el valor calculado como un número entre 0 y 10</p>
     * <p>Obsérvese que una preferencia por otra característica no deja de ser una característica también. Indica el valor que el agente da a esa característica</p>   
     * @return ArrayList con la lista de valores de las preferencias por otras características que tiene el agente
     * @see Funciones#getMultiplicador
     * @see Caracteristica
     */
    protected jade.util.leap.ArrayList iniciarPreferencias(){
    	
    	 jade.util.leap.ArrayList preferencias=new ArrayList();
    	
    	 int i=0;
    	 while (i<this.numeroCaracteristicas){
    		 String nombre="caracteristica";
    		 nombre=nombre+i+"Preferencia";
    		 float  valor=Funciones.getMultiplicador();//Le damos un valor aleatorio
    		 Caracteristica preferencia=new Caracteristica(nombre,valor);
    		 preferencias.add(preferencia);
    		 i++;
    		 
    		 //Control
    		 System.out.println(this.getLocalName()+": "+nombre+" "+valor);
    	 }
		return preferencias;
    }
    	


    
    
    
    
    /**
     * <p>Crear query</p>
     * <p>Función que crea un mensaje ACLMessenger con perfomativa QUERY_REF, protocolo FIPA_QUERY y envía como contenido: el argumento pTexto</p>
     * <p>Asigna como emisor al agente que lo ejecuta</p>
     * @param pTexto texto que se enviará en el mensaje
     * @return ACLMessage El mensaje que se envía
     * @see <a href="http://jade.tilab.com/doc/api/jade/lang/acl/ACLMessage.html">ACLMessage</a>  
     */
    private ACLMessage crearQuery(String pTexto){
    		
    		//Creamos el mensaje
    	   ACLMessage msg = new ACLMessage(ACLMessage.QUERY_REF);
    	   //Asignamos el protocolo del mensaje
    	   msg.setProtocol(FIPANames.InteractionProtocol.FIPA_QUERY);
    	 //Añadimos el lenguaje empleado
			msg.setLanguage(new SLCodec().getName());
    	   //Añadimos el emisor
    	   msg.setSender(this.getAID());			
    		//Añadimos el contenido
    	   
    		msg.setContent(pTexto);
    		
		return msg; //Retornamos el mensaje
    	
    }  
    

    
    
    
    
    /**
     * <p>Crear RequestObject </p>
     * <p>Función que crea un mensaje ACLMessenger con perfomativa REQUEST, protocolo FIPA_REQUEST y envía como contenido: el argumento o</p>
     * <p>Asigna como emisor al agente que lo ejecuta</p>
     * @param o Objeto que se enviará en el mensaje
     * @return ACLMessage El mensaje que se envía
     * @see <a href="http://jade.tilab.com/doc/api/jade/lang/acl/ACLMessage.html">ACLMessage</a>  
     */
    private ACLMessage crearRequestObject(Serializable o){
		
		//Creamos el mensaje
	   ACLMessage msg = new ACLMessage(ACLMessage.REQUEST);
	   //Asignamos el protocolo del mensaje
	   msg.setProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST);
	 //Añadimos el lenguaje empleado
		msg.setLanguage(new SLCodec().getName());
	   //Añadimos el emisor
	   msg.setSender(this.getAID());			
		//Añadimos el contenido
	   
		try {
			msg.setContentObject(o);
		} catch (IOException e) {
			System.err.println("No se ha podido crear el request");
			e.printStackTrace();
		}
		
	return msg; //Retornamos el mensaje
	
}
    
    
    
    
    
    
    
    /**
     * <p>Crear plantilla query</p>
     * <p>Crea una plantilla para escuchar mensajes con el protocolo FIPA_QUERY
     * y perfomative QUERY_REF</p>
     * @return MessageTemplate  La plantilla creada
     * @see <a href=http://jade.tilab.com/doc/api/jade/lang/acl/MessageTemplate.html>MessageTemplate</a>
     */
    private  MessageTemplate crearPlantillaQuery(){
		//Creamos la plantilla de respuesta a Query
		MessageTemplate protocolo = MessageTemplate.MatchProtocol(FIPANames.InteractionProtocol.FIPA_QUERY);
        MessageTemplate performativa = MessageTemplate.MatchPerformative(ACLMessage.QUERY_REF);
        MessageTemplate plantilla = MessageTemplate.and(protocolo, performativa);
        
        return plantilla;    
	}
   
  
    
    
    
    

 /**
  * <p>Calcula Valoración</p>
  * <p>Dada una lista de caracteristicas y otra de preferencias sobre esas características se calcula
  * la valoración de un agente con esa lista de características según la lista de preferencias del elemento que realiza la valoración</p>
  * <p>
  * El cálculo se hace sumando el producto de caracteristica1*Preferencia1+ 
  * caracteristica2*caracteristica2Preferencia+....etc...
  * <p>
  * @param caracteristicasP ArrayList Lista de caracteristicas del elemento a valorar
  * @param preferenciasP    ArrayList  Lista de las preferencias por esas características del elemento que efectúa la valoración
  * @return float. La valoración del agente con caracteristicasP realizado por el agente con preferenciasP
  */
private float calcularValoracion(ArrayList caracteristicasP,ArrayList preferenciasP){
	
	int i=0;
	float valoracion=0;
	while (i<caracteristicasP.size()){
		Caracteristica caracteristica=(Caracteristica) caracteristicasP.get(i);
		Caracteristica preferencia=(Caracteristica) preferenciasP.get(i);
		valoracion=valoracion+caracteristica.getValorCaracteristica()*preferencia.getValorCaracteristica();
		i++;
	}

	return valoracion;
	
}






/**
 * <p>getNumeroParejas</p>
 * <p>Obtiene el número de parejas que tiene el elemento</p>
 * @return int numero de parejas */

@SuppressWarnings("unused")
private int getNumeroParejas(){
	return this.numeroParejas;
}

/**
 * <p>getCaracteristicas</p>
 * <p>Obtiene un ArrayList con las características del elemento</p>
 * @return ArrayList con las características
 */
private jade.util.leap.ArrayList getCaracteristicas(){
	return this.caracteristicas;
}

/**
 * <p>getPreferencias</p>
 * Obtiene las preferencias del elemento
 * @return ArrayList con las preferencias
 */
private jade.util.leap.ArrayList getPreferencias(){
	return this.preferencias;
}



    /**
     * <p>Busqueda de emparejadores</p>
   * <p>Comportamiento que busca emparejadores que ofrezcan servicios en el DF con:<p>
   * <ul>
   * <li>name=="emparejar"</li>
   * <li>type=="parejas"</li>
   * </ul>
   * <p>Añade los AID de los emparejadores encontrados al campo <i>listaEmparejadores</i></p>
   * @see <a href=http://jade.tilab.com/doc/api/jade/core/AID.html>AID</a>
   *
   */
  private class BusquedaDeEmparejadores extends Behaviour{
  	
	  /**
		 * Constante para serializar
		 */
		private static final long serialVersionUID = -5653133673666614756L;
		/**
		 * Variable que guarda si el comportamiento está en estado de finalización o no
		 */
		boolean completado=false;//Variable que nos sirve para comprobar la finalización del comportamiento
  	@Override
  	public void action() {
  		System.out.println(myAgent.getLocalName()+" buscando emparejador"); 			
  		    	
  	
  		//Creamos el modelo sobre el que se realizarán las consultas
	    	DFAgentDescription template=new DFAgentDescription();
	    	
	    	//Seleccionamos el servicio que se proporciona y su tipo    	
	    	ServiceDescription templateSd = new ServiceDescription();    	
	    	templateSd.setName("emparejar");	
			
			//Lo añadimos al modelo de consulta
	  		template.addServices(templateSd);	
	    	
	    	try
	    	{
	    		//Creamos la consulta
	    		SearchConstraints sc = new SearchConstraints();	    
	    		// Seleccionamos cuántos resultados esperamos como máximo
	    		sc.setMaxResults(new Long(100));
	    		   			    	    	
	    	//Hacemos la búsqueda indicando al agente los criterios de selección
	    	DFAgentDescription[] result = DFService.search(myAgent, template, sc);
	    	
	    	
	    	
	    	//Si coinciden se añaden a la lista de agentes conocidos por el agente
	    	if (result.length>=1){
	    	for (int i = 0; i < result.length; ++i) {
	  				DFAgentDescription dfds = result[i];
	  			//Obtenemos el AID del agente encontrado	
				AID provider= dfds.getName();										
						
						//Añadimos a la lista los emparejadores encontrados
						listaEmparejadores.add(provider);
							
				completado=true;//Hacemos que el comportamiento pueda finalizar	
	    	}
	    	
	    	}
	    	}catch(FIPAException e){
	    		System.err.println("Error temporal en la búsqueda DF del agente: "+myAgent.getLocalName());
	    		System.err.println(e.getMessage());
	    	}
  		}
  			

  	@Override
  	public boolean done() { 
    				//Control    			  
  				if (completado){
  					System.out.println(myAgent.getLocalName()+"ha encontrado "+listaEmparejadores.size()+" agentes emparejador");
  				}
  		return completado;//Retorna la variable completado, si su valor es false el comportamiento se volverá a ejecutar
  		
  	}
  
}
  
  
  
  
  
  
  
  
  /**
   * <p>Añadir emparejadores a la lista de emparejadores</p>
   *<p>Comportamiento que añade los emparejadores existentes a la lista de emparajadores conocidos por el Elemento <i>listaEmparejadores</i></p>
   */
    private class AddEmparejadoresComoReceptor extends Behaviour	{
   
    	/**
    	 * Constante para serializar
    	 */
	private static final long serialVersionUID = 8764949162345412797L;
		/**
    	 * Variable que guarda si el comportamiento está en estado de finalización o no
    	 */
    	boolean completado=false;//Variable que nos sirve para comprobar la finalización del comportamiento
    /**
     * Mensaje que se envía
     */
    private ACLMessage msg;
    
      /**
       * Número de emparejador que se considera en cada momento
       */
  	  private int emparejadorActual;
  	  //Constructor a partir de un mensaje
  	  public AddEmparejadoresComoReceptor(ACLMessage pMsg){
  		  this.msg=pMsg;
  		  emparejadorActual=0;
  	  }
  	  @Override
  	public void action() {
  		
  		//Marcamos como receptor a cada uno de los agentes de la lista de Agentes que dan el servicio
  		AID receptor=(AID) listaEmparejadores.get(emparejadorActual);
  		  		
  			//Añadimos el receptor actual
  			msg.addReceiver(receptor);	
  			emparejadorActual++;
  						
  			//Comprobamos si se han añadido todos los emparejadores
  			completado=emparejadorActual==listaEmparejadores.size();
  			
  		 
  	}

  	@Override
  	public boolean done() {
  		
  		return completado; 
  	}
  	  
    }
    
    
    
    
    
  /**
 * <p>Iniciador de query</p>
 * <p>Iniciador que envía un mensaje e inicia los manejadores de respuesta</p>
 * <p>Si se reciben las características de un agente con un Inform valora según las preferencias de este agente al emisor
 * <p>Reordena la lista de parejas de mayor a menor valoración de las mismas</p>
 * <p>Si completa la lista de parejas busca los emparejadores existentes y les envía un request con su ElementoCompleto como contenido</p>
 * <p> Asimismo envía al gestor de agentes un mensaje(conversationId="tiempoInicioConsultas" con el tiempo en el que ha completado su lista</p>
 *@see ElementoCompleto
 *@see ListaParejas
 *@see BusquedaDeEmparejadores
 *@see AddEmparejadoresComoReceptor
 *@see InitiatorMiRequest
 *@see <a href="http://jade.tilab.com/doc/api/jade/lang/acl/ACLMessage.html">ACLMessage</a>  
 */
private class InitiatorMiQuery extends AchieveREInitiator
{
	
		
	/**
	 * Constante para serializar
	 */
	private static final long serialVersionUID = -7367529945776802469L;
	//Constructor
   		public InitiatorMiQuery (Agent miAgente,ACLMessage queryMsg) {
        super(miAgente,queryMsg);
      
	        
    }
	//Manejador del Inform
	protected void handleInform(ACLMessage msg){
		//Control
		System.out.println("Recibida respuesta de "+msg.getSender().getLocalName()+" a mi query "+myAgent.getLocalName());//AID
		try {
			
			//Recibimos el contenido del mensaje y lo guardamos en una lista.
			ArrayList caracteristicasEmisor=(ArrayList) msg.getContentObject();	
			
			//Receptor
			Elemento receptor=(Elemento) myAgent;
									
			//Calculamos la valoracion del emisor
			float emisorValoracion=calcularValoracion(caracteristicasEmisor, receptor.getPreferencias());
						
			//Buscamos la posición en la que se encuentra el emisor en nuestra lista
			int index=listaParejas.getIndexAID(msg.getSender());
			
			
			//Obtenemos el ElementoValorado de ese emisor
			ElementoValorado AgenteValoradoEmisor=(ElementoValorado) listaParejas.get(index);
			
			//Actualizamos su valoración
			AgenteValoradoEmisor.updateValoracion(emisorValoracion);
			
			//Ordenamos la lista
			listaParejas=listaParejas.ordenarMayorAMenor();
			//Actualizamos el número de valoraciones
			numeroValoraciones++;
			
			//Si el numeroValoraciones iguala el tamaño de la lista de parejas se ha terminado
			if (numeroValoraciones==listaParejas.size()){
				
				//Control
				System.out.println(myAgent.getLocalName()+" ha completado su lista");
				
				//Convertimos en Elemento
				Elemento miElemento= (Elemento) myAgent;
								
				//Guardamos en una lista los AIDs de la lista de parejas
				jade.util.leap.ArrayList listaParejas=miElemento.listaParejas.toLocalName_List();
				
				//Creamos una instancia de ElementoCompleto
				ElementoCompleto e=new ElementoCompleto(myAgent.getLocalName(), miElemento.getCaracteristicas(),miElemento.getPreferencias(),listaParejas,myAgent.getClass().toString());
				
				//Creamos una request cuyo contenido es el ElementoCompleto creado
				ACLMessage request = crearRequestObject((Serializable) e);
				
								
				//Buscamos emparejadores
				addBehaviour (new BusquedaDeEmparejadores());
				
				//Añadimos los emparejadores encontrados como receptores
				addBehaviour(new AddEmparejadoresComoReceptor(request));
				
				//Añadimos el comportamiento para enviar el request y manejar el resultado de la request
				addBehaviour(new InitiatorMiRequest(myAgent, request));
				
				
						//Enviamos un mensaje al gestor de agentes con el tiempo en el que se ha  completado su lista
				    		ACLMessage msgGest = new ACLMessage(ACLMessage.INFORM_REF);
				    		msgGest.setConversationId("tiempoInicioConsultas");
				    		//Añadimos el lenguaje empleado
							msgGest.setLanguage(new SLCodec().getName());
							//Añadimos el contenido el nombre del archivo que se ha creado
							
							long time_start = System.currentTimeMillis();
							
							try {
								msgGest.setContentObject(time_start);
							} catch (IOException e2) {
								// TODO Auto-generated catch block
								e2.printStackTrace();
							}
							
							//Añadimos el receptor
							msgGest.addReceiver(gestorAgentes);
							
							send(msgGest);
				
				
			}
			
		} catch (UnreadableException e) {
			System.err.println("No se ha podido leer el objeto recibido");
			e.printStackTrace();
		}catch (ExcepcionAgentNotFound e){
			System.err.println(e.getMessage());
			e.printStackTrace();
		} catch (Throwable e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
	}
	protected void handleNotUnderstood(ACLMessage msg){
		System.err.println("No se ha entendido mi pregunta");
	}
	protected void handleRefuse(ACLMessage msg){
		System.err.println("Mi pregunta se ha rechazado");
	}
	protected void handleFailure(ACLMessage msg){
		System.err.println("Ha habido algún error en la gestión de la respuesta a mi pregunta+\n"+msg.toString());
	}
}
    



/**
 * <p>Responder queries (enviar características)</p>
 * <p>Comportamiento para realizar las respuestas a queries de una determinada plantilla
 * Si el contenido es "dimeTusCaracteristicas" se envía como respuesta una lista con el
 * AID del agente y sus características (caracteristica1, caracteristica2, caracteristica3)
 * <p>
 *@see <a href=http://jade.tilab.com/doc/api/jade/core/AID.html>AID</a>
 *@see Elemento#getCaracteristicas()
 */
class ResponderQuery extends AchieveREResponder
{
	/**
	 * Constante para serializar
	 */
	private static final long serialVersionUID = 1425238697232017997L;


/*El constructor recibe el agente que lo ha creado 
y una plantilla con los filtros para mensajes entrantes
*/
		public ResponderQuery (Agent myAgent, MessageTemplate mt)
		{
		super (myAgent, mt);
		}
		
		
				protected ACLMessage prepareResponse (ACLMessage query) throws NotUnderstoodException
				{
					
					String contenido=query.getContent();
					ACLMessage respuesta = query.createReply();
							
							switch(contenido){
							
							/*Utilizado para controlar que se ha recibido el mensaje y que están ordenadas como se pretende
							Imprime el nombre del agente y la valoración de cada uno de los agentes de su lista*/
							case "control":
								
								jade.util.leap.Iterator it = listaParejas.iterator();
								while(it.hasNext()){
											ElementoValorado agente = (ElementoValorado) it.next();
								System.err.println(agente.getAID().getLocalName()+" :"+agente.getValoracion());				
																			
								}
							return respuesta;
							
							
							case "dimeTusCaracteristicas": 
								//Control
								System.out.println("\n"+myAgent.getLocalName()+" ha recibido petición de características del agente "+query.getSender().getLocalName());
								
																	
								try {
									Elemento agente=(Elemento) myAgent;
									
									//Añadimos el contenido del mensaje
									respuesta.setContentObject((Serializable) agente.getCaracteristicas());	
									
									//Hacemos que la respuesta sea tipo INFORM
									respuesta.setPerformative(ACLMessage.INFORM);
									
								
								} catch (Exception e) {
									System.err.println("No se han podido enviar las características del agente "+myAgent.getLocalName());
									e.printStackTrace();
									respuesta.setPerformative(ACLMessage.FAILURE);
									
								}
												
								return respuesta;							
								
								
							default: throw new NotUnderstoodException("Mensaje no entendido");
									
							}			
				}
		
}
/**
 * <p>Iniciador de request</p>
 * <p>Envía un mensaje y gestiona las respuestas</p>
 * <p> Solo dará mensajes textuales ya que no se requieren más acciones.</p>
 */
public class InitiatorMiRequest extends AchieveREInitiator{

	/**
	 * Constante para serializar
	 */
	private static final long serialVersionUID = 7136920551903940725L;
		public InitiatorMiRequest(Agent a, ACLMessage msg) {
		super(a, msg);	
		
			}
		
		protected void handleAgree(ACLMessage msg){			
			System.out.println("Agree recibido de "+msg.getSender().getLocalName()+" para "+myAgent.getLocalName());
		}
		protected void handleInform(ACLMessage msg){			
			System.err.println("Informe recibido "+msg.getSender()+" : "+msg.getContent());
		}
		
		protected void handleNotUnderstood(ACLMessage msg){
			System.err.println("No se ha entendido mi request");
		}
		protected void handleRefuse(ACLMessage msg){
			System.err.println("Mi request se ha rechazado");
		}
		protected void handleFailure(ACLMessage msg){
			System.err.println("Ha habido algún error en la gestión de la respuesta a mi request");
		}	
}

/**
 * <p>Escuchar Objetos (listaParejas)</p>
 * <p>Comportamiento que escucha mensajes con un determinado ConversationId</p>
 * <p>Si se recibe un mensaje con ConversationId="listaParejas" forma con él
 * su propia lista de parejas <i>listaParejas</i><p/>
 * <p>Bloquea la siguiente ejecución hasta la recepción de otro mensaje</p>
 */
private class EscucharObjetoConversation extends Behaviour{
	
	/**
	 * Constante para serializar
	 */
	private static final long serialVersionUID = 2048214500804421319L;
	/**
	 * Identificador de conversación en el que escucha el comportamiento
	 */
	private String conversationId;
	
	/**
	 * Constructor del comportamiento, escuchará las conversaciones de identificador pConversation
	 * @param pConversation
	 */
	public EscucharObjetoConversation(String pConversation){
		this.conversationId=pConversation;
	}
	@Override
	public void action() {
		//Creamos un filtro, solo se recibirán los mensajes cuyo identificador de conversacion coincida con conversationId
		
		MessageTemplate mt=MessageTemplate.MatchConversationId(conversationId);
		//Recepción no bloqueante de un mensaje según el filtro creado mt
		ACLMessage msg = myAgent.receive(mt);			
				
	if (msg != null) {
	
					if (conversationId.equals("listaParejas")){
						try {
							
							ArrayList listaParejasAID =(ArrayList) msg.getContentObject();
							
							jade.util.leap.Iterator it = listaParejasAID.iterator();
							
							while (it.hasNext()){
								
								ElementoValorado pareja=new ElementoValorado((AID) it.next(),(float)0.0);
								listaParejas.add(pareja);
							}
							
						} 
						
						catch (UnreadableException e) {
							
							System.err.println("El mensaje recibido no es una lista de parejas\n" +e.getMessage());
						}
	}
		}else
		{
		//Bloqueamos el comportamiento hasta recibir un mensaje en la cola
		block();
		}
  }

			@Override
			public boolean done() {
				// TODO Auto-generated method stub
				return false;
			}
  
}

/**
 * <p>Envia un mensaje a la lista de parejas</p>
 * <p>Comportamiento que añade como destinatarios de un mensaje determinado a todos los miembros de su <i>listaParejas</i><p>
 * <p>Cuando ha completado su lista envía el mensaje añadiendo el comportamiento <i>InitiatorMiQuery</i></p>
 *@see InitiatorMiQuery
 */
private class EnviarMensajeAParejas extends Behaviour	{
		
	/**
	 * Constante para serializar
	 */
	private static final long serialVersionUID = 3329564502452438085L;
		/**
		 * Variable que guarda si el comportamiento está en estado de finalización o no
		 */
		boolean completado=false;
		  private ACLMessage msg;
		  
		  //Constructor a partir de un mensaje
		  public EnviarMensajeAParejas(ACLMessage pMsg){
			  this.msg=pMsg;
		  }
		  @Override
		public void action() {
		
			  if (listaParejas.size()==numeroParejas){
			//Marcamos como receptor a cada uno de los agentes de la lista de Agentes que dan el servicio
			ElementoValorado receptor=(ElementoValorado) listaParejas.get(agentesPreguntados);
			AID receptorAID=receptor.getAID();
						
				msg.addReceiver(receptorAID);	
				agentesPreguntados++;
							
				//Comprobamos si se han añadido todos los agentes, si es así el comportamiento puede finalizar
				completado=agentesPreguntados==listaParejas.size();
			  }
			  else completado=false;
		}
		
		@Override
		public boolean done() {
			if (completado){
				
				//Envía el mensaje e inicia los manejadores de respuestas.
				addBehaviour(new InitiatorMiQuery(myAgent, msg));
				
				

			}
			return completado; 
		}
		  
		}
}


  

	   
   

   

