package tfgMarriageStable;

import jade.content.lang.sl.SLCodec;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.OneShotBehaviour;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPANames;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.FailureException;
import jade.domain.FIPAAgentManagement.NotUnderstoodException;
import jade.domain.FIPAAgentManagement.RefuseException;
import jade.domain.FIPAAgentManagement.SearchConstraints;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.lang.acl.UnreadableException;
import jade.proto.AchieveREInitiator;
import jade.proto.AchieveREResponder;
import jade.util.leap.ArrayList;
import jade.util.leap.HashMap;
import jade.util.leap.Iterator;
import jade.util.leap.Map;


/**
 * <p>Agente que implementa el algoritmo de matrimonio estable</p>
 * <p>Ofrece el servicio de name="emparejar" type="parejas" owner="TFG"</p>
 * <p>Escucha Requests con protocolo FIPA_REQUEST y perfomativa REQUEST para recibir los datos de cada Elemento</p>
 * <p>Cuando termina un emparejamiento envía un mensaje al Gestor de Agentes para que continúe el programa y queda a la espera de nuevos mensajes</p>
 * <p>Si recibe un Refuse no hace nada y sigue a la espera de nuevas iteraciones</p>
 * <p>Si recibe un Agree guarda los datos en un archivo y envía el nombre de ese archivo a los agentes que ofrecen servicio de tipo "clasificacion" como Weka</p>
 * @author Alberto Hueso Alonso
 *@since 06/12/2017
 *@see GestorAgentes
 *@see ElementoCompleto
 *@see Weka
 */
public abstract class Emparejador extends Agent{
	
	/**
	 * Constante para serializar
	 */
	private static final long serialVersionUID = -3480952699225058649L;
	/**
	 * Nombre del archivo donde se guardarán los datos por defecto
	 */
	static final String NOMBRE_ARCHIVO_DATOS_POR_DEFECTO="Datos.csv";
	/**
	 * El propio agente
	 */
	private Agent esteAgente=this;
	/**
	 * Número de parejas a unir
	 */
	private int numeroParejas;//Número de parejas recibido al crear desde el manager
	
	/**
	 * AID del gestor de agentes
	 */
	private AID gestorAgentes;
	/**
	 * Guarda todos los ElementoCompleto de todas las iteraciones
	 * la clave del mapa es el localname y el valor el ElementoCompleto
	 * 
	 */
	private Map elementosCompletos=new HashMap();
	
	/**
	 * Elementos X recibidos, la clave del mapa es el localname del elemento y el valor su lista de favoritos
	 */
	 protected Map elementosX = new HashMap();
	 /**
	  * Elementos Y recibidos, la clave del mapa es el localname y el valor el valor de su lista de favoritos
	  */
	 protected Map elementosY = new HashMap();
	 /**
	  * <p>Guarda los emparejados estables durante cada iteración</p>
	  * <p>La clave es el localname de un elemento de la pareja (Y) y el valor el local name del elemento (X).</p> 
	  */
	 protected Map emparejados=new HashMap(); 
	 
	 /**
	  *<p>Guarda las parejas estables  de todas las iteraciones<p>
	  * <p>La clave es el localname de un elemento de la pareja (Y) y el valor el local name del elemento (X).</p> 
	  */
	 private Map emparejadosFinal=new HashMap();
	 
	 /**
	  * Guarda las parejas inestables de cada iteración
	  */
	 protected ArrayList noEmparejados=new ArrayList(); 
	 /**
	  * Guarda las parejas no estables de todas las iteraciones
	  */
	 private ArrayList noEmparejadosFinal=new ArrayList();
	 
	 /**
	  * Indica si hay que crear archivo de muestras de las listas de preferencias recibidas
	  */
	 private boolean filePruebas;
	/**
	 * Indica si los resultados se han de guardar en el archivo por defecto o en un nuevo archivo 
	 */
	private boolean newfile;
	
	/**
	 * Iteración en curso
	 */
	protected int iteracion=0;
	
	/**
	 * Nombre del archivo con el resultado
	 */
	private String nameFile="";
	
	/**
	 * Guarda si las preferencias son moduladas o no
	 */
	private boolean moduladas;
	
	
	
	
	
	/**
	  * <p>Registra el emparejador</p>
	  * <p>Inicializa el número de agentes recibidos y recupera como argumentos el número de parejas, el AID del gestor de agentes, si hay que crear archivos de pruebas y si hay que  guardar los datos del emparejamiento en un archivo nuevo o en el archivo por defecto NOMBRE_ARCHIVO_DATOS_POR_DEFECTO</p>
	  *<p>Abre un comportamiento para escuchar los requests de los elementos con sus ElementosCompletos</p>
	  *@see #crearPlantillaRequest
	  *@see ManejadorResponderRequests
	  */
	 public void setup(){
		 
		//Registramos el agente en el Domain Facilitator
		 RegistrarEmparejadorDF();
		 
			 
		//Recuperamos los argumentos que se le pasan al inicio
		 numeroParejas=(int) getArguments()[0];
		 gestorAgentes=(AID) getArguments()[1];
		 filePruebas=(boolean) getArguments()[2];
		 newfile=(boolean) getArguments()[3];
		 moduladas=(boolean) getArguments()[4];
		 
		//Registramos el lenguaje SL en el agente
		   	Funciones.registrarLenguajeSL(this);  
				 
		 //Creamos una plantilla Request para escuchar Requests
		 MessageTemplate plantilla = crearPlantillaRequest();
		 
		 addBehaviour(new ManejadorResponderRequests(this, plantilla));
	}
	 
	 
	/**
	 * Función de limpieza al eliminar el agente
	 */
	public void takeDown(){
		deRegistrarDF();
		
		System.out.println("El emparejador ha sido destruido");
	}
	
	
	
	
	
	
	   /**
     * <p>Registro del emparejador en el Domain Facilitator</p>
     * <p>Asigna el nombre "emparejar", el tipo "parejasEmparejar" y el propietario "TFG"</p>
     * 
     */
   private void RegistrarEmparejadorDF(){
    	
    	
    	//Petición de registro en el DF
    	DFAgentDescription dfd=new DFAgentDescription();
    	
    	//Servicio que se proporciona
    	ServiceDescription sd= new ServiceDescription();  
    	
    	//Asignamos el nombre al servicio
    	sd.setName("emparejar");
    	
    	
    	//Se añade el tipo y el propietario
    	sd.setType("parejasEmparejar");
    	sd.setOwnership("TFG");
    	
    	
    	//Añadimos los servicios al descriptor de agentes
    	dfd.addServices(sd);
    	dfd.setName(getAID());
    	
    	try {
    		//Añadimos al registro
			DFService.register(this, dfd);
			System.out.println("Se ha registrado el agente "+getLocalName());
		} catch (FIPAException e) {				
			
				System.err.println("No se ha podido registrar el agente "+getLocalName());
				
				
						
		}
    	
    }
   /**
    * Desregistra el agente del Domain Facilitator
    */
	 private void deRegistrarDF(){
	    	
		    try {
		    	//Solicitamos el desregistro
				DFService.deregister(this);
				System.out.println("El agente "+getLocalName()+" se ha desregistrado");
			} catch (FIPAException e) {
				
				System.err.println("No se ha podido desregistrar el agente");
			}
		    }
	 
	 /**
	  * Comprueba si el número de ElementosX y el número de ElementosY son iguales y si son iguales al número de parejas
	  * @return boolean con true o false, según sea cierto o falso
	  */
	 private boolean checkNumeroParejas(){
		 if (elementosX.size()==elementosY.size()&&elementosX.size()==numeroParejas){
			 
			 return true;
		 }else 
			 {
			 return false;
			 }
	 }
	 
	
	 /**
	  * <p>Crea unMensaje Request que solicita al gestor agentes que continue el programa porque el emparejador ha terminado una iteración</p>
	  * 	  * @return ACL con perfomative REQUEST, protocolo FIPA_REQUEST, contenido "continueProgram" y destinatario el gestor de agentes
	  */
	 private ACLMessage nextIteracion(){
		//Creamos el mensaje
		   ACLMessage msg = new ACLMessage(ACLMessage.REQUEST);
		   //Asignamos el protocolo del mensaje
		   msg.setProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST);
		   msg.setPerformative(ACLMessage.REQUEST);
		 //Añadimos el lenguaje empleado
			msg.setLanguage(new SLCodec().getName());
			msg.setContent("continueProgram");
		   //Añadimos el emisor
		  
		   msg.addReceiver(gestorAgentes);
		  
		 
		   
		 		 return msg;
	 }
	 
	 
	 
	 
	 
	 /**
	  * Función abstracta, se implementará en las clases hijos, con un algoritmo diferente en cada caso
	  * @return Map Un mapa en el que un elemento será la clave y la pareja el valor
	  */
	  protected  Map emparejar(){
		  System.out.println("Emparejando");
		  return null;
	  }
	  
	  
	  
	  
	  /**
	   * Obtiene la cabecera del archivo de resultado con las características y preferencias
	   * @return String con nombres de características y preferencias
	   */
	  private String getCabecera(){
		  String cabecera = "Pareja,";
		  
		  //Iterador para recorrer la lista de claves de elementosCompletos
		  Iterator it = elementosCompletos.keySet().iterator();
		  
		  //Obtenemos el nombre del primer elemento		  
		  String nombreElemento =(String) it.next();
		  
		  //Obtenemos ese ElementoCompleto
		  ElementoCompleto unElementoCompleto=(ElementoCompleto) elementosCompletos.get(nombreElemento);
		  
		  //Obtenemos el número de características de ese elemento completo
		  int numCaracteristicas=unElementoCompleto.getCaracteristicas().size();
		 
		 //Recorremos las características y se va creando la cabecera
		 int i=0;
		 while (i<numCaracteristicas){
			 cabecera=cabecera.concat("Xcaracteristica"+i+",");
			 cabecera=cabecera.concat("Xpreferencia"+i+",");
			 i++;
		 }
		 
		 //A continuación añadimos las preferencias
		 int j=0;	 
		 while (j<numCaracteristicas){
			 cabecera=cabecera.concat("Ycaracteristica"+j+",");
			 cabecera=cabecera.concat("Ypreferencia"+j+",");
			 j++;
		 }
		 
		 //Añadimos el final de la línea
		 cabecera=cabecera.concat("clase"+"\r\n");
		
		//Retornamos la cabecera completa
		return cabecera;
		  
	  }
	  
	  
	  
	  
	  
	  /**
	   * <p>Obtiene una línea a partir de dos nombres de ElementosCompletos</p>
	   * <p>Se forma añadiendo el valor de una característica del elemento, 
	   * seguida de su preferencia por esa característica hasta completar el elementoX. 
	   * A continuación se repite con el elementoY 
	   * Para terminar se añade la clase.
	   * Cada característica se separa con ","
	   * @param X String nombre de un elemento X
	   * @param Y String nombre de un elemento Y
	   * @param clase String que dice si la pareja es estable o no estable
	   * @return String con una línea
	   * @see ElementoCompleto
	   */
	  private String getLinea(String X,String Y,String clase){
		 
		 //Primer elemento de la línea 
		 String linea=X+" : "+Y+",";
		 
		 //Obtenemos los elementosCompletos
		 ElementoCompleto elementoX = (ElementoCompleto) elementosCompletos.get(X);
		 ElementoCompleto elementoY = (ElementoCompleto) elementosCompletos.get(Y);
		 
		 //Obtenemos las características y preferencias de cada uno de ellos
		 
		 ArrayList xCaracteristicas=elementoX.getCaracteristicas();
		 ArrayList xPreferencias=elementoX.getPreferencias();
		 ArrayList yCaracteristicas=elementoY.getCaracteristicas();
		 ArrayList yPreferencias=elementoY.getPreferencias();
		 
		 //Recorremos y añadimos a la línea las características y preferencias del elemento X
		 int i=0;	 
		 while (i<xCaracteristicas.size()){
			 Caracteristica caracteristica=(Caracteristica) xCaracteristicas.get(i);
			 Caracteristica preferencia=(Caracteristica) xPreferencias.get(i);
			 
			 linea=linea.concat(caracteristica.getValorCaracteristica()+","+preferencia.getValorCaracteristica()+",");
			 i++;
		 }
		 //Recorremos y añadimos a la línea las características y preferencias del elemento Y
		 i=0;
		 while (i<yCaracteristicas.size()){
			 Caracteristica caracteristica=(Caracteristica)yCaracteristicas.get(i);
			 Caracteristica preferencia=(Caracteristica) yPreferencias.get(i);
			 
			 linea=linea.concat(caracteristica.getValorCaracteristica()+","+preferencia.getValorCaracteristica()+",");
			 i++;
		 }
		 
		 //Añadimos el fin de línea
		 linea=linea.concat(clase+"\r\n");
		  return linea;
	  }
	  
	  
	  
	  
	 /**
		 * Crea una plantilla MessageTemplate para protocolo FIPA_REQUEST y Perfomative REQUEST.
		 * @return MessageTemplate 
		 * @see <a href=http://jade.tilab.com/doc/api/jade/lang/acl/MessageTemplate.html>MessageTemplate</a>
		 */
	  private  MessageTemplate crearPlantillaRequest(){
			
			//Creamos la plantilla de respuesta a REQUEST
			MessageTemplate protocolo = MessageTemplate.MatchProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST);
	        MessageTemplate performativa = MessageTemplate.MatchPerformative(ACLMessage.REQUEST);
	        MessageTemplate plantilla = MessageTemplate.and(protocolo, performativa);
	        
	        return plantilla;
	    
		}
		
		
		
		

		/** 
		 * Función que guarda un mapa con los elementos de un grupo y sus preferencias en un fichero de texto
		 * @param mapa
		 * @param nombreFichero
		 */
		// Basado en  https://blogdeaitor.wordpress.com/2008/11/21/java-escribir-en-fichero/
		private void storeFileTest(Map mapa, String nombreFichero){
		    //Creamos nuestro String que vamos a escribir en el fichero
		    String texto="";
		    		      	      
		      //Recorremos las claves del mapa
		      Iterator it = mapa.keySet().iterator();
		      
		      while (it.hasNext()){
		    	  String agente = (String) it.next();//Recuperamos el agente del mapa
		    	  texto=texto.concat(agente+":");//Iniciamos la línea con el nombre del agente y :
		    	  System.out.println("\nEscribiendo en archivo las preferencias del agente: "+agente);
		    	  
		    	  //Recuperamos la lista de preferencias del agente
		    	  ArrayList lista =(ArrayList) mapa.get(agente);
					int i=0;
					
					//Recorremos la lista y añadimos cada pareja preferida seguida de una coma
					while (i<lista.size()){
					
						String pareja = (String) lista.get(i);
						if (i!=lista.size()-1){
						texto=texto.concat(pareja+",");
						}
						else{
							 //Añadimos la última pareja del agente y hacemos un salto de carro y de línea
							texto=texto.concat(pareja+"\r\n");
						}
					i++;
					}
		      }  
		      
		      //Creamos el archivo
		    Funciones.createFile(texto, nombreFichero,false);
		  }
		    
		  
		
		/**
		 * <p>Manejador de respuestas a Requests.</p>
		 * <p>Muestra un mensaje de recepción</p>
		 * <p>Si el contenido es un objeto ElementoCompleto acepta la petición (En caso contrario se produce una excepción UnreadableException)
		 * y lo añade a la lista de elementosX o elementosY según la clase a la que pertenezcan</p>
		 * <p>Cuando ambas listas están completas (esto es,tienen el número de parejas esperados) los empareja</p>
		 * <p>Además, si filePruebas==true, guarda las listas de preferencias recibidas en dos archivos, elementosX.out y elementosY.out</p>
		 */
	private	class ManejadorResponderRequests extends AchieveREResponder
	    {
		/**
		 * Constante para serializar
		 */
			private static final long serialVersionUID = -3191849805021280360L;
			/**
			 * Agente que inicia el manejador
			 */
			private Agent agent;

			public ManejadorResponderRequests(Agent a,MessageTemplate mt) {
	            super(a,mt);
	            this.agent=a;
	        }
	 
	        protected ACLMessage handleRequest(ACLMessage request)throws NotUnderstoodException, RefuseException
	        {
	        	Emparejador emparejador=(Emparejador) agent;
				
	         System.out.println(myAgent.getLocalName()+ " ha recibido un mensaje tipo Request de "+request.getSender().getLocalName()+ "Tenemos: "+emparejador.elementosCompletosSize());
	            
	         
	        	 try {
	        		 //Recuperamos el contenido del mensaje
	        		 ElementoCompleto elemento = (ElementoCompleto) request.getContentObject();
	        		
	        		 //Creamos la respuesta
	        		 ACLMessage agree = request.createReply();
	                    agree.setPerformative(ACLMessage.AGREE);//Perfomativa AGREE
	                  
						agree.setContent("Recibidos los datos");//Marcamos el contenido del mensaje
						
						//Guardamos el elemento completo en la lista de elementos recibidos
						elementosCompletos.put(elemento.getAgent(), elemento);
	                 
						String clase=elemento.getClase();
						
						//Comprobamos si es de la clase ElementoX o de la clase ElementoY
						if (clase.contains("ElementoX")){
							clase="ElementoX";
							elementosX.put(elemento.getAgent(), elemento.getListaParejas());
						}
						if (clase.contains("ElementoY")){
							clase="ElementoY";
							elementosY.put(elemento.getAgent(), elemento.getListaParejas());
						}	
						
						//Si no es un ElementoX o ElementoY se arroja una excepción
						if (!(clase.equals("ElementoX")||clase.equals("ElementoY"))){
							 throw new RefuseException("No es un elemento emparejable");
						}
						
						//Cuando se tienen todos los elementos esperados se emparejan
						if (checkNumeroParejas()){						
							
							addBehaviour(new Emparejar());
							iteracion++;
							
							//Si filePruebas==true se guardan las listas de preferencias en dos archivos
							if (filePruebas){
								//Guardamos las listas de las preferencias de ambos grupos en dos archivos						
								storeFileTest(elementosX,"elementosX.out");
								storeFileTest(elementosY,"elementosY.out");
								System.out.println("Archivos de muestras guardados: Xstring.out y Ystring.out  ");
							}
						}
						
						
						return agree; 
						
				} catch (UnreadableException e) {					
					e.printStackTrace();
					throw new NotUnderstoodException("No se ha entendido el mensaje");	
				}               
				
	            
	       
				}      

			//No realiza ninguna acción
	        protected ACLMessage prepareResultNotification(ACLMessage request,ACLMessage response) throws FailureException{
				return null;
	        	
	        }
	            
	    }
		
		/**
		 * <p>Comportamiento que en primer lugar muestra las preferencias de los dos grupos</p>
		 * <p>Después las empareja según la función <i>emparejar</i></p>
		 * <p>Muestra las parejas estables y las parejas no estables producidas</p> 
		 * <p>Reinicia las listas temporales de cada iteracion</p>
		 *<p>Abre un InitiatorMiRequest solicitando al gestor de agentes que borre los elementos de los dos grupos</p>
		 @see #emparejar
		 @see Funciones#verPreferencias(Map, Map)
		 @see Funciones#verParejas(Map, String)
		 @see Funciones#verParejas(ArrayList, String)
		 @see Emparejador#nextIteracion()
		 */
		private class Emparejar extends OneShotBehaviour{
			/**
			 * Constante para serializar
			 */
			private static final long serialVersionUID = 7040485388686049922L;

			@Override
			public void action() {
				//Control
				
				Funciones.verPreferencias(elementosX,elementosY);
				
					if (checkNumeroParejas()){
				 emparejados=emparejar();
				 
				 //Unimos los emparejados en el que guarda el resultado de todas las iteraciones
				 emparejadosFinal=Funciones.joinMaps(emparejadosFinal, emparejados);
				 
				 //Unimos los no emparajados al que guarda los resultados finales
				 noEmparejadosFinal=Funciones.joinArrays(noEmparejadosFinal, noEmparejados);
				 
				//Control 
				Funciones.verParejas(emparejados,"SON PAREJAS ESTABLES\n");
				Funciones.verParejas(noEmparejados,"SON PAREJAS INESTABLES\n");
			
				//Reiniciamos las listas de la iteración
				elementosX = new HashMap(); 
				elementosY = new HashMap(); 
				emparejados=new HashMap(); 					 
				noEmparejados=new ArrayList();
				 
				//Mensaje para Request continueProgram para el gestor de agentes
				ACLMessage msg = nextIteracion();
				//Se añade el comportamiento que envía el mensaje y gestiona las respuestas
				addBehaviour(new InitiatorMiRequest(esteAgente,msg));
					}
			}
			
		}
		
		/**
		 * 
		 * <p>Crear un archivo con el resultado</p>
		 * <p>Comportamiento que guarda en un archivo los resultados</p>
		 * <p>Si el valor de newfile==true crea un nuevo archivo, en caso contrario sobreescribe el archivo NOMBRE_ARCHIVO_DATOS_POR_DEFECTO</p>
		 * <p>Este archivo tendrá una cabecera con los nombres de las características y preferencias</p>
		 * <p>Después, cada línea estará formada por los valores de las características y preferencias de los miembros
		 * de las parejas y la clase a la que pertenecen, estable o no"</p>
		 * <p>Cuando ha terminado de escribirlo ejecuta la acción EnviarNombreArchivoResultadoClasificadores()</p>
		 *	<p> Por último borra el agente</p>
		 *@see Emparejador#getCabecera
		 *@see Emparejador#getLinea(String, String, String)
		 *@see Emparejador#EnviarNombreArchivoResultadoClasificadores()
		 */
		private class ArchivoResultado extends OneShotBehaviour {

			/**
			 * Constante para serializar
			 */
			private static final long serialVersionUID = 2644299240065336641L;

			@Override
			public void action() {
				String fechaHora=Funciones.getDate();
				nameFile=NOMBRE_ARCHIVO_DATOS_POR_DEFECTO;
				
				//Si se ha marcado newfile se crea un nuevo archivo
				if (newfile){
				  nameFile = esteAgente.getLocalName()+"_parejas_"+numeroParejas+"_Iteraciones_"+iteracion+"_"+fechaHora+"modulacion_"+moduladas+".csv";
				}
				String cabecera=getCabecera();
				String texto=cabecera;
				
				
				//Añadimos las líneas de las parejas estables
				Iterator it=emparejadosFinal.keySet().iterator();
				while (it.hasNext()){
					String Y=(String) it.next();
					String X=(String) emparejadosFinal.get(Y);
					String linea=getLinea(X,Y,"Estable");
					System.out.println("Escribiendo Pareja "+X+","+Y);
					texto=texto.concat(linea);				
				}
				
				//Añadimos las líneas de las parejas no estables encontradas
				it=noEmparejadosFinal.iterator();
			    while (it.hasNext()){
			    	Pareja noPareja=(Pareja) it.next();
			    	String Y=noPareja.getY();
			    	String X=noPareja.getX();
			    	System.out.println("Escribiendo Pareja "+X+","+Y);
			    	String linea=getLinea(X,Y,"NoEstable");
			    	texto=texto.concat(linea);
			    }
			
			//Creamos el archivo
			Funciones.createFile(texto, nameFile,false);
			
			//Enviamos el nombre del archivo de resultado a los Clasificadores
			EnviarNombreArchivoResultadoClasificadores();    
				doDelete();//Borramos el agente
				
			}
			
		}
		
		
		
		/**
		 * <p>Iniciador de Request</p>
		 * <p>Envía un mensaje (debería ser un request) y muestra un mensaje según las diferentes posibles respuestas</p>
		 * *<p> Si se recibe un Agree además inicia el comportamiento ArchivoResultado</p>
		 * @see ArchivoResultado
		 */
		private class InitiatorMiRequest extends AchieveREInitiator{
			
			/**
			 * Constante para serializar
			 */
			private static final long serialVersionUID = -9087346811102432854L;
				public InitiatorMiRequest(Agent a, ACLMessage msg) {
				super(a, msg);	
				
					}
				
				protected void handleAgree(ACLMessage msg){	
					
					System.out.println("Agree recibido de "+msg.getSender().getLocalName()+" para "+myAgent.getLocalName());
					//Añadimos el comportamiento de crear el archivo con el resultado
					addBehaviour(new ArchivoResultado());
				}
				protected void handleInform(ACLMessage msg){			
					System.err.println("Informe recibido "+msg.getSender()+" : "+msg.getContent());
				}
				
				protected void handleNotUnderstood(ACLMessage msg){
					System.err.println("No se ha entendido mi request");
				}
				protected void handleRefuse(ACLMessage msg){
					System.out.println("Mi request se ha rechazado");
				}
				protected void handleFailure(ACLMessage msg){
					System.err.println("Ha habido algún error en la gestión de la respuesta a mi request");
				}	
		}


		/**
		 * Obtiene el numero de elementos del campo elementosCompletos
		 * @return int número de elementos
		 */
		private int elementosCompletosSize() {
			int size = this.elementosCompletos.size();
			
			return size;
		}	
		
		/**
		 * <p>Envía el nombre del archivo de resultados <i>nameFile</i> a los agentes Weka que ofrezcan servicios de tipo "clasificacion"</p>
		 *<p>El mensaje se envía como un INFORM con ConversationId "archivoDatos"</p>
		 *@see Weka
		 *@see <a href=http://jade.tilab.com/doc/api/jade/lang/acl/ACLMessage.html>ACL Message</a>
		 */
		private void EnviarNombreArchivoResultadoClasificadores() {
						
	    		//Creamos el modelo sobre el que se realizarán las consultas
		    	DFAgentDescription template=new DFAgentDescription();
		    	
		    	//Servicio que se proporciona	    	
		    	ServiceDescription templateSd = new ServiceDescription();
		    	
		    	//Tipo de servicio 		    	
				templateSd.setType("clasificacion");
		  		template.addServices(templateSd);	

		    	try
		    	{
		    		ACLMessage msg = new ACLMessage(ACLMessage.INFORM_REF);
		    		msg.setConversationId("archivoDatos");
		    		//Añadimos el lenguaje empleado
					msg.setLanguage(new SLCodec().getName());
					//Añadimos el contenido el nombre del archivo que se ha creado
					msg.setContent(nameFile);
					
		    		//Creamos la consulta
		    		SearchConstraints sc = new SearchConstraints();
		    		
		    				    	
		    		   			    	    	
		    	//Hacemos la búsqueda indicando al agente los criterios de selección
		    	DFAgentDescription[] result = DFService.search(esteAgente, template, sc);
		    	
		    	for (int i = 0; i < result.length; ++i) {
	  				DFAgentDescription dfds = result[i];
	  			//Obtenemos el AID del agente encontrado	
				AID providerAID = dfds.getName();
				
				msg.addReceiver(providerAID);
				
		    	}
		    	//Enviamos el mensaje
		    	send(msg);
		    	System.out.println("Mensaje con nombre de archivo de resultados enviado a clasificadores");
			}catch(FIPAException e){
				System.err.println("No se ha podido enviar ningún mensaje a clasificadores");
				
			}
		}
		
		
}