package tfgMarriageStable;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.Random;
import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.classifiers.functions.Logistic;
import weka.core.Instances;
import weka.core.converters.CSVLoader;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.Normalize;
import weka.filters.unsupervised.attribute.Remove;
import weka.filters.unsupervised.instance.Randomize;

/**
 * Clase que se encarga de realizar el estudio de un nombre y ruta de archivo de training pasado como argumento al crear el agente o recibida como mensaje desde un Emparejador, y opcionalmente un archivo de test pasado como argumento</p>
 * <p>Sobreescribe el método UtilizarDatos de <Clasificador>
 * <p>Utiliza para el estudio el clasificador Logistic</p>
 * @author Alberto Hueso Alonso
 * @since 13-12-2017
 *@see #UtilizarDatos
 *@see Emparejador
 *@see Clasificador
 */
public class Weka extends Clasificador {

/**
	 * Constante para Serialización
	 */
	private static final long serialVersionUID = -3288952114770081389L;
/**
 * Guarda las instancias del set de entrenamiento
 */
private Instances trainingSet=null;
/**
 * Guarda las instancias del set de test
 */
private Instances testData=null;

/**
 * Guarda si se ha recibido un archivo de test
 */
private boolean testFile=false;

/**
 * Guarda el nombre del archivo de test
 */
private String testDataFileName=null;

/**
 * Guarda el tiempo de inicio de una evaluación
 */
private long time_start;

/**
 * Guarda el tiempo de finalización de una evaluación
 */
private long time_end;





	/**
	 * <p>Recupera los argumentos que se le pasan al crearlo (nombre del archivo de training y, opcionalmente, del archivo de test</p
	 *<p>Si hay archivo de test, lo carga y lo preprocesa</p>
	 *<p>Utiliza los datos de ambos archivos según la función UtilizarDatos
	 *@see #UtilizarDatos()
	 *@see #loadFile(String)
	 *@see #preProceso(Instances) 
	 */
	public void setup(){
		super.setup();	
		
		System.out.println("Iniciado Weka");
		
		
		//Recuperamos los argumentos introducidos
		Object[] args = getArguments();
		
		//Intentamos obtener el  archivo de test, el segundo argumento es su nombre
		try {
			
			//Obtenemos el segundo argumento, el archivo de test
			 testDataFileName = (String) args[1];
			
			 //Cargamos el archivo
			testData=loadFile(testDataFileName);
			
			//Lo preprocesamos
			testData=preProceso(testData);
			
			System.out.println("Archivo de test");
			
			//Mostramos el resumen de los datos
			mostrarResumen(testData);
			
			//Actualizamos valor de testFile
			testFile=true;
		
		 //Si ha habido algún error no se ha podido leer el archivo
		}catch (Exception e){
			System.out.println("No se ha obtenido archivo de test");
		}
		
		//Función principal con todas las acciones
		UtilizarDatos();
		
		
	}
	
	/**
	 * <p>Carga un archivo arff o csv de nombre nameFile</p>
	 * @param nameFile
	 * @return Instances con los datos del archivo cargados
	 * @throws Exception 
	 * @see <a href=http://weka.sourceforge.net/doc.stable/weka/core/Instances.html>Instances</a>
	 * @see <a href=http://weka.sourceforge.net/doc.dev/weka/core/converters/CSVLoader.html>CSVLoader</a>
	 */
	private Instances loadFile(String nameFile) throws Exception{
		
		Instances dataSet=null;
				
		//Si es un archivo CSV cargamos los datos con CSVLoader
		if (nameFile.endsWith(".csv")||nameFile.endsWith(".CSV")){
			System.out.println(nameFile);
			
		//Creamos el cargador de CSV	
	    CSVLoader loader = new CSVLoader();
	    //Añadimos el origen del archivo
	    loader.setSource(new File(nameFile));
	    
	    //Cargamos el set de Instances
	     dataSet = loader.getDataSet();
	    
		}
		
		//Si es un archivo ARFF utilizamos new Instances
		if (nameFile.endsWith(".arff")|| nameFile.endsWith(".ARFF")){
		
			
			 dataSet = new Instances(new BufferedReader(new FileReader(nameFile)));	
		}
		
		//Si no se ha cargado el dataSet se muestra un mensaje de error	
		if (dataSet==null){
				
			System.err.println("\n\nHa habido algún problema.No se ha podido leer el archivo de datos"+nameFile+"\nCompruebe el tipo de archivo\n\n");
		}
		
		//Retornamos el dataSet obtenido
		return dataSet;
	}
	
	/**
	 * <p>Carga el archivo de training y, si existe, el archivo de test</p>
	 * <p>A partir del archivo de training crea un clasificador Logistic</p>
	 * <p>Si existe el archivo de test evalúa este según el clasificador creado con el conjunto de training </p>
	 * <p>Si no existe el archivo de test, evalúa el archivo de training con el clasificador mediante una validación cruzada estratificada</p>
	 * <p>Muestra los resultados</p>
	 * <p>Cierra el sistema</p>
	 * <p>Si no hay archivo de training queda a la espera de recibirlo (desde un emparejador)</p>
	 * @see #mostrarResumen(Instances)
	 * @see #mostrarResultados(Evaluation)
	 * @see #clasificarLogistic(Instances)
	 * @see #evaluarCrossValidation
	 * @see #evaluar(Classifier, Instances, Evaluation)
	 * @see Emparejador
	 * @see Funciones#cerrarSistema(jade.core.Agent)
	 */
	protected void UtilizarDatos(){
		System.out.println("Leeremos el archivo enviado por un emparejador "+nameFile);
		
		try {			
			//Cargamos el archivo
			trainingSet=loadFile(nameFile);	
			
			} catch (Exception e) {
					System.err.println("Ha habido algún problema al leer el archivo: "+nameFile+" "+e.getMessage());
				}
		
			
		if (trainingSet!=null){
			
			trainingSet=preProceso(trainingSet);
			
			System.out.println("Archivo de entrenamiento");
			mostrarResumen(trainingSet);
			
			//Creamos un clasificador
			Logistic clasificador=clasificarLogistic(trainingSet);
			
			try {
			
				System.out.println("Evaluando el clasificador "+clasificador.getClass().getCanonicalName());
				
				//Iniciamos el contador de tiempo dela evaluación			
				time_start= System.currentTimeMillis();
			
			//Solo hay archivo de test por lo que se efectúa una validación cruzada	
			if (!testFile){
				System.out.println("No hay archivo de test, se efectuará una validación cruzada sobre el mismo archivo");
				testData=trainingSet;
				testData=shuffle(testData);
				
				Evaluation	eval = new Evaluation(testData);
				
				
				 evaluarCrossValidation(clasificador,testData,eval);
				//Finalizamos el contador de tiempo		
				  time_end = System.currentTimeMillis();
				  
				  mostrarResultados(eval);
			}
			
			//Evaluamos el archivo de test y training
			else {				
				Evaluation	eval = new Evaluation(testData);
				System.out.println("Archivos de test y training diferentes");
				evaluar(clasificador,testData,eval);
				//Finalizamos el contador de tiempo		
				  time_end = System.currentTimeMillis();
				  mostrarResultados(eval);
				
			}
				
		}catch (Exception e) {
				System.err.println("No se ha podido evaluar");
				e.printStackTrace();
			}
			
			System.out.print("\n\n\nFIN\n\n\n");
			
			//Cerramos el sistema
			Funciones.cerrarSistema(this);
			
			
		}else 
		
		//Espera mensajes de un emparejador
		System.out.println("\nEsperando mensajes");
		
		} 
		
/**
 * <p>Aplica el filtro Remove al conjunto de instancias pDataSet eliminando el atributo i-ésimo (contando de 1-N)</p>
 * @param i String número de columna que se eliminará
 * @param pDataSet
 * @return Instances El mismo conjunto pDataSet pero sin la columna número i
 * @see <a href=http://weka.sourceforge.net/doc.stable/weka/filters/unsupervised/attribute/Remove.html>Remove</a>
 * @see <a href=http://weka.sourceforge.net/doc.stable/weka/core/Instances.html>Instances</a> 
 */
	private Instances quitarFila(int i,Instances pDataSet) {
		
			
		try{
			
			//Obtenemos el nombre del atributo. Se resta uno porque el índice de los atributos empieza en 0
			String nombre=pDataSet.attribute(i-1).name();	
			
			System.out.println("\nFiltrando el atributo: "+nombre);
			
			String pColumna=String.valueOf(i);
				
				//Marcamos las opciones del filtro -R indica que la opción marcará el rango y pColumna, la columna específica
				String[] opts = new String[]{ "-R", pColumna};
				
				//Creamos el objeto filtro Remove
				Remove remove = new Remove();
				
				//Añadimos las opciones
				remove.setOptions(opts);
				
				//Pasamos el trainingSet al filtro
				remove.setInputFormat(pDataSet);
				
				//Aplicamos el filtro
				Instances newData = Filter.useFilter(pDataSet, remove);
				
			return newData;
				
			}
			catch (Exception e){
				System.err.println("No se ha podido filtrar la fila número: "+i);
				e.printStackTrace();
			}
			
			return pDataSet;
	}
	/**
	 * Muestra un resumen de data
	 * @param data
	 * @see <a href=http://weka.sourceforge.net/doc.stable/weka/core/Instances.html>Instances</a> 
	 */
	private void mostrarResumen(Instances data){
		//Mostramos el resumen de los datos cargados
		System.out.println("\nRESUMEN:");
		System.out.println(data.toSummaryString()+"\n\n\n");
	}
	/**
	 * Aplica un filtro de normalización al set de Instances pDataSet
	 * @param pDataSet
	 * @return Instances el set de Instances normalizado
	 * @see <a href=http://weka.sourceforge.net/doc.stable/weka/core/Instances.html>Instances</a>
	 * @see <a href=http://weka.sourceforge.net/doc.stable/weka/filters/unsupervised/attribute/Normalize.html>Normalize</a>
	 */
	private Instances normalizar(Instances pDataSet){
		
		System.out.println("\nNormalizando");
		try{
		Normalize normalize=new Normalize();
		normalize.setInputFormat(pDataSet);
		Instances newData=Filter.useFilter(pDataSet, normalize);
		
		return newData;
		} catch (Exception e){
			System.err.println("No se ha podido normalizar los datos");
			e.printStackTrace();
		}
		
		return pDataSet;
		
	}
	

	/**
	 * Crea un clasificador Logistic a partir de un set de datos DataSet
	 * @param dataset
	 * @return Logistic clasificador Logistic
	 * @see <a href=http://weka.sourceforge.net/doc.stable/weka/core/Instances.html>Instances</a>
	 * @see <a href=http://weka.sourceforge.net/doc.stable/>Logistic</a>
	 */
	private Logistic clasificarLogistic(Instances dataset){
		
		//Creamos el clasificador Logistic
		Logistic clasificador = new Logistic();
		
		try {
			//Construye el clasificador a partir de los datos
			clasificador.buildClassifier(dataset);
			
			//Retorna el clasificador
			return clasificador;
			
		} catch (Exception e) {
			System.out.println("No se ha podido construir el clasificador" +clasificador.getClass().getName());
			e.printStackTrace();
		}
		return null;
	}
	/**
	 * Desordena un set de Instancias
	 * @param pDataSet El conjunto de instancias
	 * @return Instances El conjunto desordenado
	 * @see <a href=http://weka.sourceforge.net/doc.stable/weka/filters/unsupervised/instance/Randomize.html>Randomize</a>
	 * @see <a href=http://weka.sourceforge.net/doc.stable/weka/core/Instances.html>Instances</a>
	 */
	private Instances shuffle(Instances pDataSet){
		
			
		//Creamos el filtro
		Randomize randomize=new Randomize();	
		
		try {
						
			//Introducimos los datos
			randomize.setInputFormat(pDataSet);
			
			//Aplicamos el filtro
			Instances newData = Filter.useFilter(pDataSet, randomize);
		
			//Retornamos el nuevo valor
		return newData;
			
		} catch (Exception e) {
			System.err.println("No se ha podido desordenar el dataset");
			e.printStackTrace();
		}
		//Retornamos el valor original
		return pDataSet;
		
	}
	
	/**
	 * Evalúa con un clasificador (pClasificador) un test de datos (pTestData) a partir de un Evaluation (pEval)
	 * <p>La evaluación se realiza con validación cruzada estratificada de diez iteraciones
	 * @param pClasificador
	 * @param pTestData
	 * @param pEval
	 * @throws Exception
	 * @see <a href=http://weka.sourceforge.net/doc.stable/weka/core/Instances.html>Instances</a>
	 * @see <a href=http://weka.sourceforge.net/doc.dev/weka/classifiers/Classifier.html>Classifier</a>
	 * @see <a href=http://weka.sourceforge.net/doc.stable/weka/classifiers/Evaluation.html>Evaluation</a>
	 *
	 */
	private void evaluarCrossValidation(Classifier pClasificador,Instances pTestData,Evaluation pEval) throws Exception{
					
		//Creamos el random
		Random rand = new Random(1);
		
		//Ajustamos el número de iteraciones
		int folds = 10;
		
		  
		  //Efectuamos la evaluación
		pEval.crossValidateModel(pClasificador, pTestData, folds, rand);
				
	}
	
	/**
	 * Evalúa con un clasificador (pClasificador) un conjunto de test (pTestData), y un Evaluation (pEval)
	 * @param pClasificador
	 * @param pTestData
	 * @param pEval
	 * @throws Exception
	 * @see <a href=http://weka.sourceforge.net/doc.stable/weka/core/Instances.html>Instances</a>
	 * @see <a href=http://weka.sourceforge.net/doc.dev/weka/classifiers/Classifier.html>Classifier</a>
	 * @see <a href=http://weka.sourceforge.net/doc.stable/weka/classifiers/Evaluation.html>Evaluation</a>
	 */
	private void evaluar(Classifier pClasificador,Instances pTestData,Evaluation pEval) throws Exception{
		
						
			System.out.println("Evaluando el clasificador "+pClasificador.getClass().getCanonicalName());
			
			  
			//Efectuamos la evaluación sobre pTestData
			pEval.evaluateModel(pClasificador, pTestData);
			
			
		
	}
	
	/**
	 * Preprocesa un conjunto de Instancias
	 * <p>Quita la primera fila (las parejas) y aplica el filtro de normalización</p>
	 * @param pDataSet
	 * @return Instances . El DataSet sin la primera fila y normalizados los datos
	 * @see #quitarFila(int, Instances)
	 * @see #normalizar(Instances)
	 */
	private Instances preProceso(Instances pDataSet){
		//Filtramos la primera fila (las parejas)
		pDataSet=quitarFila(1,pDataSet);
		
		//Normalizamos los datos			
		pDataSet=normalizar(pDataSet);
		
		//Asignamos la clase		
		pDataSet.setClassIndex(pDataSet.numAttributes()-1);
		return pDataSet;
	
		
	}
	/**
	 * Muestra los resultados de una Evaluación y, si el nombre tiene la estructura adecuada [ver función obtenerParejasIteracionesCompletoModulacion(String)]  los guarda en el archivo "resultados.csv"
	 * <ul>Los datos que muestra son:
		 * 	<li>Resumen</li>
		 *  <li>Matriz de confusión</li>
		 *  <li>Tiempo que ha tardado en milisegundos</li>
		 *	<li>Archivo de entrenamiento</li>
		 *	<li>Archivo de test</li>
		 *	<li>Aciertos</li>
		 *	<li>Errores</li>
		 *	<li>Precision</li>
		 *	<li>Porcentaje error</li>
		 *	<li>Tasa true positives Estables</li>
		 *	<li>Tasa true positives No Estables</li>
		 *	<li>Precision Estables</li>
		 *	<li>Precision No Estables</li>
		 *	<li>F-Measure Estables</li>
		 *	<li>F-Measure NoEstables</li>
		 *	<li>Índice Kappa</li>
	 * </ul>
	 * @param eval Evaluation el evaluador del que se mostrarán los resultados y guardará los datos en el archivo "resultados.csv"
	 * @throws Exception
	 * @see <a href=http://weka.sourceforge.net/doc.stable/weka/classifiers/Evaluation.html>Evaluation</a>
	 * @see Funciones#createFile(String, String, Boolean)
	 * @see #obtenerParejasIteracionesCompletoModulacion(String)
	 */
	private void mostrarResultados(Evaluation eval) throws Exception{
		System.out.println(eval.toSummaryString("Evaluation results:\n",false));
		System.out.println(eval.toMatrixString());
				
				
		System.out.println("El evaluador ha tardado: "+(time_end-time_start)+" milisegundos");
		
		//Marcaremos 2 decimales y utilizaremos el punto como marcador de decimales
		DecimalFormatSymbols simbolos = new DecimalFormatSymbols();
		simbolos.setDecimalSeparator('.');
		DecimalFormat df = new DecimalFormat("0.00",simbolos);
		
		
		double aciertos = eval.correct();
		double errores=eval.incorrect();
		String precision=df.format((1-eval.errorRate())*100)+"%";		
		String tasaErrores=df.format(eval.errorRate()*100)+"%";
		String fMeasureEstables = df.format(eval.fMeasure(0));
		String fMeasureNoEstables=df.format(eval.fMeasure(1));
		String establesTruePositiveRate=df.format(eval.truePositiveRate(0));
		String noEstablesTruePositiveRate=df.format(eval.truePositiveRate(1));		
		String establesPrecision=df.format(eval.precision(0));
		String noEstablesPrecision=df.format(eval.precision(1));
		String kappa=df.format(eval.kappa());
		
		System.out.println("Archivo de entrenamiento:"+nameFile);
		System.out.println("Archivo de test:"+testDataFileName);
		System.out.println("Aciertos:"+aciertos);
		System.out.println("Errores:"+errores);
		System.out.println("Precision:"+precision);
		System.out.println("Porcentaje error:"+tasaErrores);		
		System.out.println("Tasa true positives Estables:"+establesTruePositiveRate);
		System.out.println("Tasa true positives No Estables:"+noEstablesTruePositiveRate);
		//System.out.println("Tasa false positive Estables"+establesFalsePositiveRate);
		//System.out.println("Tasa true positive No Estables"+noEstablesFalsePositiveRate);
		System.out.println("Precision Estables:"+establesPrecision);
		System.out.println("Precision No Estables:"+noEstablesPrecision);
		System.out.println("F-Measure Estables:"+fMeasureEstables);
		System.out.println("F-Measure NoEstables:"+fMeasureNoEstables);
		System.out.println("Kappa estadístico:"+kappa);
		
		//Datos de parejas, iteraciones, completitud y modulación
		String training="null,null,null,null";
		String test="null,null,null,null";
		
		//Si el nombre del archivo de training es convertible se actualiza training
		try{
			training=obtenerParejasIteracionesCompletoModulacion(nameFile);
			
			//Si el nombre del archivo de test es convertible se actualiza test
			if (testDataFileName!=null){
				test=obtenerParejasIteracionesCompletoModulacion(testDataFileName);
			}
					}
		//Si no es convertible el nombre del archivo no se hace nada
		catch (Exception e){
			//No se produce ninguna acción
		}

		//Creamos el texto para guardar en archivo
		String texto=nameFile+","+testDataFileName+","+
					 training+","+test+","+
					(time_end-time_start)+","+aciertos+","+
					errores+","+precision+","+tasaErrores+","+
					establesTruePositiveRate+","+noEstablesTruePositiveRate+","+					
					establesPrecision+","+noEstablesPrecision+","+
					fMeasureEstables+","+fMeasureNoEstables+","+kappa+"\r\n";	
		
		//Guardamos en un archivo
		Funciones.createFile(texto, "resultados.csv", true);
		
	}
	
	/**
	 * <p>Método que a partir de un nombre archivo retorna un String con el numero de parejas, el número de iteraciones, si el algoritmo utilizado ha sido completo y si se ha modulado, separados por comas</p>. 
	 * <p>Para ello el archivo ha de contener una cadena "parejas_" seguida de un entero, y una cadena "Iteraciones_" seguida de un número. En caso contrario arroja una excepción</p>
	 * <p>Ejemplo: A partir de GaleShapleyCompleto_parejas_100_Iteraciones_1_2017_12_13_10_07_24modulacion:false se retornaría
	 * <p>100,1,Completo,false</p>
	 * <p>
	 * @param nombreArchivo String nombre del archivo
	 * @return String con cadena en la que separadas por comas encontramos el número de parejas, el número de iteraciones, Completo o NoCompleto y true o false. Ejemplo: "100,1,Completo,false"
	 * @throws NumberFormatException
	 */
	private String obtenerParejasIteracionesCompletoModulacion(String nombreArchivo) throws NumberFormatException, IndexOutOfBoundsException{
	
		String numeroParejas;
		String numeroIteraciones;
		
			//Separamos en la palabra parejas
			String[] parejas = nombreArchivo.split("parejas");
			//Volvemos a separar con el _ y obtenemos el número de parejas
			numeroParejas = (parejas[1].split("_"))[1];
			 
			//Convertimos en entero. Si no es un entero arrojará una excepción
			 @SuppressWarnings("unused")
			int numero=Integer.parseInt(numeroParejas);

			 //Procedemos de forma similar con Iteraciones
			 String[] iteraciones= nombreArchivo.split("Iteraciones");
			 numeroIteraciones=(iteraciones[1].split("_"))[1];		 
			 numero=Integer.parseInt(numeroIteraciones);
		
			 
		 String completitud;
		 if (nombreArchivo.contains("Completo")){
			 completitud="Completo";
		 }else{
			 completitud="NoCompleto";
		 }
		
		 String modulacion;
		 
		 if (nombreArchivo.contains("true")){
			 modulacion="true";
		 }else {
			 modulacion="false";
		 }
		 
		 //Formamos la cadena de resultado
		String resultado=numeroParejas+","+numeroIteraciones+","+completitud+","+modulacion;
		 
		return resultado;
	}
	
}
